/* eslint-disable id-length */
import { objectClean } from '@dzeio/object-util'
import { StandardSchemaV1 } from '@standard-schema/spec'
import Schema, { isNull, parceable, SchemaNullable } from './Schema'
import { SchemaJSON, ValidationError, ValidationResult } from './types'
export default abstract class SchemaItem<Type = any> implements StandardSchemaV1<Type> {

	// standard Schema V1 spec
	public '~standard': StandardSchemaV1.Props<Type> = {
		vendor: 'aptatio',
		version: 1,
		validate: (value: unknown) => {
			const res = this.parse(value)
			if (!res.valid) {
				return {
					issues: res.errors
				}
			}
			return {
				value: res.object,
				issues: res.errors,
			}
		}
	}

	/**
	 * keep public ?
	 */
	public validations: Array<{
		fn: (input: Type) => boolean
		error?: string | undefined
	}> = []

	/**
	 * Function calls saved for serialization
	 */
	public savedCalls: Array<{ name: string, args: Array<string> | IArguments | undefined }> = []


	/**
	 * Pre process the variable for various reasons
	 *
	 * It will execute every pre-process sequentially in order of being added
	 *
	 * note: The type of the final Pre-Process MUST be valid
	 */
	// eslint-disable-next-line @typescript-eslint/no-redundant-type-constituents
	public preProcess: Array<(input: unknown) => Type | unknown> = []

	/**
	 * post process the variable after it being validated
	 *
	 * it will execute each post-processes sequentially
	 *
	 * note: the type of the final post-process MUST be valid
	 */
	public postProcess: Array<(input: Type) => Type> = []

	/**
	 * list of attributes for custom works
	 */
	public readonly attributes: Array<string> = []

	private readonly items?: Array<unknown>

	private invalidError = 'the field is invalid'

	public constructor(items?: Array<unknown> | IArguments) {
		if (items && items.length > 0) {
			this.items = Array.isArray(items) ? items : Array.from(items)
		}
	}

	@parceable()
	public defaultValue(value: Type, strict = true) {
		this.addPreProcess((input) => {
			if (strict && isNull(input) || !strict && !input) {
				return value
			}
			return input
		})

		return this
	}

	/**
	* make sure the value is one of the `values`
	* @param values the values the item MUST contains
	*/
	@parceable()
	public in(...values: Array<Type>): this {
		this.addValidation((input) => values.includes(input))

		return this
	}

	public attrs(...attributes: Array<string>) {
		this.attributes.push(...attributes)
		return this
	}

	public attr(...attributes: Array<string>) {
		return this.attrs(...attributes)
	}

	public setInvalidError(err: string): this {
		this.invalidError = err
		return this
	}

	public clone(): this {
		return Schema.fromJSON(this.toJSON()) as this
	}

	public nullable(): SchemaNullable<this> {
		return new SchemaNullable(this)
	}

	/**
	 * schemas implementing unwrap can return their child component (mostly the value)
	 *
	 * ex: s.record(s.number(), s.string()) returns s.string()
	 *
	 * es2: s.nullable(s.string()) returns s.string()
	 */
	public unwrap?(): SchemaItem

	public parse(input: unknown, options?: { fast?: boolean }): ValidationResult<Type> {
		// pre process the variable
		for (const preProcess of this.preProcess) {
			input = preProcess(input)
		}

		// validate that the pre process handled correctly the variable
		if (!this.isOfType(input)) {
			return {
				valid: false,
				errors: [{ message: 'invalid type' }]
			}
		}

		// run validation checks
		const errors: Array<ValidationError> = []
		for (const validation of this.validations) {
			if (!validation.fn(input)) {

				// add the error to the list
				errors.push({
					message: validation.error ?? this.invalidError
				})

				// if the system should be fast, stop checking for other errors
				// eslint-disable-next-line max-depth
				if (options?.fast) {
					return {
						valid: false,
						object: input as (Type extends object ? Partial<Type> : Type) | undefined,
						errors: errors
					}
				}
			}
		}

		// return with errors
		if (errors.length > 0) {
			return {
				valid: false,
				object: input as (Type extends object ? Partial<Type> : Type) | undefined,
				errors: errors
			}
		}

		// post-process the variable
		for (const postProcess of this.postProcess) {
			input = postProcess(input as Type)
		}

		return {
			valid: true,
			object: input as Type
		}
	}

	public toJSON(): SchemaJSON {

		// build the JSON
		const res = {
			i: this.constructor.name,
			a: this.attributes.length > 0 ? this.attributes : undefined,
			c: this.items?.map((it) => it instanceof SchemaItem ? it.toJSON() : it),
			f: this.savedCalls
				.map((it) => (it.args ? { n: it.name, a: Array.from(it.args) } : { n: it.name }))
		}

		// cleanup the object from undefined to make it smaller
		objectClean(res, { deep: false })

		return res
	}

	protected addValidation(fn: ((input: Type) => boolean) | { fn: (input: Type) => boolean, error?: string }, error?: string) {

		this.validations.push(typeof fn === 'function' ? { fn, error } : fn)

		return this
	}

	// eslint-disable-next-line @typescript-eslint/no-redundant-type-constituents
	protected addPreProcess(fn: (input: unknown) => Type | unknown) {
		this.preProcess.push(fn)

		return this
	}

	protected addPostProcess(fn: (input: Type) => Type) {
		this.postProcess.push(fn)

		return this
	}

	public abstract isOfType(input: unknown): input is Type
}
